# coding=utf-8

from os.path import join

from flask import current_app, Blueprint, render_template, _app_ctx_stack
from werkzeug.local import LocalProxy

from guniflask.config import settings
from guniflask.utils.template import template_folder
from guniflask.utils.config import walk_modules

static_folder = join(template_folder, 'static')

apidoc = Blueprint('apidoc', __name__, url_prefix='/_apidoc',
                   template_folder=join(template_folder, 'apidoc'),
                   static_folder=static_folder,
                   static_url_path='/static')

api_doc = LocalProxy(lambda: current_app.extensions['apidoc'])


@apidoc.route('/', methods=['GET'])
def get_api_modules():
    api_module_keys = _get_api_module_keys()
    return render_template('api_modules.html', api_module_keys=api_module_keys, api_modules=api_doc.api_modules,
                           project_name=settings['project_name'])


def _get_api_module_keys():
    api_modules = api_doc.api_modules
    keys = [i for i in api_modules.keys()]
    keys.sort(key=lambda x: api_modules[x]['url_prefix'])
    return keys


@apidoc.route('/<module>', methods=['GET'])
def get_api_list(module):
    api_list = _get_api_list(module)
    return render_template('api_list.html', api_list=api_list, module=module, project_name=settings['project_name'])


def _get_api_list(module):
    api_set = api_doc.api_set
    api_list = []
    for k, v in api_set.items():
        if k.startswith(module + '.'):
            api_list.append(v)
    api_list.sort(key=lambda x: x['path'])
    return api_list


class ApiDoc:
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.register_blueprint(apidoc)
        app.extensions['apidoc'] = self

    @property
    def api_modules(self):
        ctx = _app_ctx_stack.top
        if ctx is not None:
            if not hasattr(ctx, 'api_modules'):
                self.make_api_docs()
            return ctx.api_modules

    @property
    def api_set(self):
        ctx = _app_ctx_stack.top
        if ctx is not None:
            if not hasattr(ctx, 'api_set'):
                self.make_api_docs()
            return ctx.api_set

    def make_api_docs(self):
        app = current_app
        ctx = _app_ctx_stack.top
        api_modules = {}
        api = {}

        for module in walk_modules(app.name):
            for obj in vars(module).values():
                if isinstance(obj, Blueprint):
                    doc = module.__doc__
                    if doc:
                        doc = doc.rstrip()
                    api_modules[obj.name] = dict(name=obj.name, url_prefix=obj.url_prefix, doc=doc)
        rules = {}
        for i in app.url_map.iter_rules():
            rules[i.endpoint] = i
        for e, r in rules.items():
            f = app.view_functions[e]
            name = ' '.join([i.title() for i in f.__name__.split('_')])
            methods = [i for i in r.methods if i not in ['HEAD', 'OPTIONS']]
            methods.sort()
            methods = ', '.join(methods)
            doc = f.__doc__
            if doc:
                doc = doc.rstrip()
            api[e] = dict(name=name, endpoint=e, path=r.rule, methods=methods, doc=doc)

        ctx.api_modules = api_modules
        ctx.api_set = api



# coding=utf-8

import os
from os.path import exists, join, abspath, isdir, basename, dirname
from shutil import ignore_patterns
import signal
import json

from guniflask.errors import AbortedError, TemplateError
from guniflask.utils.template import string_lowercase_underscore, string_lowercase_hyphen, jinja2_env, template_folder
from guniflask.utils.cli import readchar
from guniflask.utils.security import generate_jwt_secret
from guniflask import __version__


class Command:
    def __init__(self):
        self.exitcode = 0

    @property
    def name(self):
        return ""

    @property
    def syntax(self):
        return ""

    @property
    def short_desc(self):
        return ""

    @property
    def long_desc(self):
        return self.short_desc

    def add_arguments(self, parser):
        pass

    def process_arguments(self, args):
        pass

    def run(self, args):
        raise NotImplementedError


class Step:
    desc = None

    def __init__(self):
        self.value = None
        self.title = self.desc
        self.tooltip = None

    def process_arguments(self, settings):
        pass

    def process_user_input(self):
        while True:
            self.show_question()
            user_input = self.get_user_input()
            if self.check_user_input(user_input):
                self.show_decision()
                break
            else:
                self.show_invalid_tooltip()

    def show_question(self):
        raise NotImplementedError

    def get_user_input(self):
        raise NotImplementedError

    def show_decision(self):
        raise NotImplementedError

    def show_invalid_tooltip(self):
        raise NotImplementedError

    def check_user_input(self, user_input):
        return True

    def update_settings(self, settings):
        raise NotImplementedError

    def question(self):
        return '\033[32m?\033[0m {} \033[37m({})\033[0m'.format(self.title, self.tooltip)

    def decision(self):
        return '\033[32m?\033[0m {} \033[36m{}\033[0m'.format(self.title, self.value)

    def go_back_lines(self, n=1):
        return '\033[1A\r\033[K' * n

    def clean_line(self):
        return '\r\033[K'


class InputStep(Step):
    def show_question(self):
        print(self.question() + ' ', end='', flush=True)

    def get_user_input(self):
        return input()

    def show_decision(self):
        print(self.clean_line(), end='', flush=True)
        print(self.go_back_lines(), end='', flush=True)
        print(self.decision(), flush=True)

    def show_invalid_tooltip(self):
        print(self.go_back_lines(), end='', flush=True)


class ChoiceStep(Step):
    def __init__(self):
        super().__init__()
        self.selected = 0
        self.choices = []
        self.values = []

    def show_question(self):
        print(self.question() + ' ', end='', flush=True)
        for i in range(len(self.choices)):
            print(flush=True)
            if i == self.selected:
                print('\033[36m>\033[0m \033[36m{}\033[0m'.format(self.choices[i]), end='', flush=True)
            else:
                print('  {}'.format(self.choices[i]), end='', flush=True)
        self.hide_cursor()

    def get_user_input(self):
        ch = readchar()
        if ch == '\x03':
            self.show_cursor()
            raise KeyboardInterrupt
        if ch == '\x1a':
            self.show_cursor()
            os.kill(os.getpid(), signal.SIGSTOP)
        return ch

    def check_user_input(self, user_input):
        if user_input == '\r' or user_input == '\n':
            self.value = self.choices[self.selected]
            self.show_cursor()
            return True
        if user_input == '\x1b':
            assert readchar() == '['
            ch = readchar()
            if ch == 'A':
                self.selected -= 1
                if self.selected < 0:
                    self.selected = len(self.choices) - 1
            elif ch == 'B':
                self.selected += 1
                if self.selected >= len(self.choices):
                    self.selected = 0
        return False

    def show_decision(self):
        print(self.clean_line(), end='', flush=True)
        print(self.go_back_lines(len(self.choices)), end='', flush=True)
        print(self.decision(), flush=True)

    def show_invalid_tooltip(self):
        print(self.clean_line(), end='', flush=True)
        print(self.go_back_lines(len(self.choices)), end='', flush=True)

    def add_choice(self, desc=None, value=None):
        self.choices.append(desc)
        self.values.append(value)

    @property
    def selected_value(self):
        return self.values[self.selected]

    def show_cursor(self):
        print('\033[?25h', end='', flush=True)

    def hide_cursor(self):
        print('\033[?25l', end='', flush=True)


class BaseNameStep(InputStep):
    desc = 'What is the base name of your application?'

    def process_arguments(self, settings):
        project_dir = settings['project_dir']
        self.tooltip = string_lowercase_underscore(basename(project_dir))
        old_settings = settings['old_settings']
        if old_settings and 'project_name' in old_settings:
            self.tooltip = old_settings['project_name']

    def check_user_input(self, user_input):
        project_basename = string_lowercase_underscore(user_input)
        if not project_basename:
            project_basename = self.tooltip
        if project_basename and not str.isdigit(project_basename[0]):
            self.value = project_basename
            return True
        return False

    def show_invalid_tooltip(self):
        print('\033[31m>>\033[0m Please input a valid project name', end='', flush=True)
        super().show_invalid_tooltip()

    def update_settings(self, settings):
        settings['project_name'] = self.value


class PortStep(InputStep):
    desc = 'Would you like to run your application on which port?'

    def process_arguments(self, settings):
        self.tooltip = 8000
        old_settings = settings['old_settings']
        if old_settings and 'port' in old_settings:
            self.tooltip = old_settings['port']

    def check_user_input(self, user_input):
        user_input = user_input.strip()
        if user_input:
            port = self.parse_port(user_input)
        else:
            port = self.parse_port(self.tooltip)
        if port is not None:
            self.value = port
            return True
        return False

    def show_invalid_tooltip(self):
        print('\033[31m>>\033[0m Please input a valid port (0 ~ 65535)', end='', flush=True)
        super().show_invalid_tooltip()

    def update_settings(self, settings):
        settings['port'] = self.value

    @staticmethod
    def parse_port(port):
        try:
            res = int(port)
        except (ValueError, TypeError):
            pass
        else:
            if 0 <= res < 65536:
                return res


class AuthenticationTypeStep(ChoiceStep):
    desc = 'Which type of authentication would you like to use?'

    def __init__(self):
        super().__init__()
        self.tooltip = 'Use arrow keys'
        self.add_choice('JWT authentication', 'jwt')

    def update_settings(self, settings):
        security = self.selected_value
        settings['authentication_type'] = security
        if self.selected_value == 'jwt':
            settings['jwt_secret'] = generate_jwt_secret()


class ConflictFileStep(InputStep):
    def __init__(self, file):
        super().__init__()
        self.title = 'Overwrite {}?'.format(file)
        self.tooltip = 'Y/n/a/x'

    def check_user_input(self, user_input):
        user_input = user_input.strip().lower()
        if not user_input:
            user_input = 'y'
        if len(user_input) > 1 or user_input not in 'ynax':
            return False
        self.value = user_input
        return True

    def show_invalid_tooltip(self):
        print('\033[31m>>\033[0m Please enter a valid command', end='', flush=True)
        super().show_invalid_tooltip()


class InitCommand(Command):
    @property
    def syntax(self):
        return '[options]'

    @property
    def name(self):
        return 'init'

    @property
    def short_desc(self):
        return 'Initialize a project'

    def add_arguments(self, parser):
        parser.add_argument('-d', '--root-dir', dest='root_dir', metavar='DIR',
                            help='application root directory')
        parser.add_argument('-f', '--force', dest='force', action='store_true', default=False,
                            help='force generating an application')

    def run(self, args):
        project_dir = abspath(args.root_dir or '')
        self.print_welcome(project_dir)
        try:
            init_json_file = join(project_dir, '.guniflask-init.json')
            old_settings = None
            try:
                with open(init_json_file, 'r', encoding='utf-8') as f:
                    old_settings = json.load(f)
                if args.force:
                    raise FileNotFoundError
                settings = old_settings
                self.print_regenerate_project()
            except (FileNotFoundError, json.JSONDecodeError):
                settings = self.get_settings_by_steps(project_dir, old_settings=old_settings)
                with open(init_json_file, 'w', encoding='utf-8') as f:
                    json.dump(settings, f, indent=2, sort_keys=True)
            self.copy_files(project_dir, settings)
        except (KeyboardInterrupt, AbortedError):
            print(flush=True)
            self.print_aborted_error()
            self.exitcode = 1

    def get_settings_by_steps(self, project_dir, old_settings=None):
        steps = [BaseNameStep(), PortStep(), AuthenticationTypeStep()]
        cur_step, total_steps = 1, len(steps)
        settings = {'project_dir': project_dir,
                    'old_settings': old_settings,
                    'guniflask_version': __version__}
        print(flush=True)
        for step in steps:
            step.title = '({}/{}) {}'.format(cur_step, total_steps, step.title)
            step.process_arguments(settings)
            step.process_user_input()
            step.update_settings(settings)
            cur_step += 1
        del settings['project_dir']
        del settings['old_settings']
        return settings

    def copy_files(self, project_dir, settings):
        settings = dict(settings)
        project_name = settings['project_name']
        settings['project_dir'] = project_dir
        settings['project__name'] = string_lowercase_hyphen(project_name)

        print(flush=True)
        self.print_copying_files()
        self.force = False
        self.copytree(join(template_folder, 'project'), project_dir, settings)
        print(flush=True)
        self.print_success()

    def copytree(self, src, dst, settings):
        names = os.listdir(src)
        ignored_names = ignore_patterns('*.pyc')(src, names)
        for name in names:
            if name in ignored_names:
                continue
            src_path = join(src, name)
            dst_name = self.render_string(name, **settings)
            dst_path = join(dst, dst_name)
            dst_rel_path = self.relative_path(dst_path, settings['project_dir'])

            if isdir(src_path):
                self.copytree(src_path, dst_path, settings)
            else:
                content = self.read_file(src_path)
                if dst_name.endswith('.tmpl'):
                    dst_name = dst_name[:-5]
                    dst_path = join(dst, dst_name)
                    dst_rel_path = self.relative_path(dst_path, settings['project_dir'])
                    try:
                        content = self.render_string(content, **settings)
                    except TemplateError:
                        continue
                if exists(dst_path):
                    raw = self.read_file(dst_path)
                    if content == raw:
                        self.print_copying_file('identical', dst_rel_path)
                    else:
                        if self.force:
                            self.print_copying_file('force', dst_rel_path)
                            self.write_file(dst_path, content)
                        else:
                            self.print_copying_file('conflict', dst_rel_path)
                            cf_step = ConflictFileStep(dst_rel_path)
                            cf_step.process_user_input()
                            user_input = cf_step.value
                            if user_input == 'y' or user_input == 'a':
                                self.print_copying_file('force', dst_rel_path)
                                self.write_file(dst_path, content)
                                if user_input == 'a':
                                    self.force = True
                            elif user_input == 'n':
                                self.print_copying_file('skip', dst_rel_path)
                            elif user_input == 'x':
                                raise AbortedError
                else:
                    self.print_copying_file('create', dst_rel_path)
                    self.write_file(dst_path, content)

    @staticmethod
    def relative_path(path, fpath):
        path = abspath(path)
        fpath = abspath(fpath)
        if path.startswith(fpath):
            path = path[len(fpath):]
            if path.startswith(os.path.sep):
                path = path[1:]
        return path

    @staticmethod
    def read_file(path):
        with open(path, 'r', encoding='utf-8') as f:
            return f.read()

    @staticmethod
    def write_file(path, raw):
        d = dirname(path)
        if not exists(d):
            os.makedirs(d)
        with open(path, 'w', encoding='utf-8') as f:
            f.write(raw)

    @staticmethod
    def render_string(raw, **kwargs):
        env = jinja2_env()
        return env.from_string(raw).render(**kwargs)

    @staticmethod
    def print_welcome(project_dir):
        print('\033[37mWelcome to gunicorn generator\033[0m \033[33mv{}\033[0m'
              .format(__version__), flush=True)
        print('\033[37mApplication file will be created in folder:\033[0m \033[33m{}\033[0m'
              .format(project_dir), flush=True)

    @staticmethod
    def print_regenerate_project():
        print('\033[32mThis is an existing project, using the configuration from .guniflask-init.json '
              'to regenerate the project...\033[0m')

    @staticmethod
    def print_success():
        print('\033[32mApplication is created successfully.\033[0m', flush=True)

    @staticmethod
    def print_aborted_error():
        print('\033[33mProcess is aborted by user.\033[0m', flush=True)

    @staticmethod
    def print_copying_files():
        print('Copying files:', flush=True)

    @staticmethod
    def print_copying_file(t, path):
        color = 0
        if t == 'identical':
            color = 36
        elif t == 'conflict':
            color = 31
        elif t == 'create':
            color = 32
        elif t == 'force' or t == 'skip':
            color = 33
        print('\033[{}m{:>9}\033[0m {}'.format(color, t, path), flush=True)


class VersionCommand(Command):
    @property
    def name(self):
        return "version"

    @property
    def short_desc(self):
        return "Print the version"

    def run(self, args):
        print("guniflask version {}".format(__version__))



# coding=utf-8

import os
from os.path import join, exists
import subprocess
import time
from threading import Thread
import json

from guniflask import __version__


def show_version():
    res = subprocess.run("guniflask version", shell=True)
    assert res.returncode == 0


def init_project(proj_dir):
    settings = {
        'guniflask_version': __version__,
        'authentication_type': None,
        'port': 8000,
        'project_name': 'foo'
    }
    with open(join(proj_dir, '.guniflask-init.json'), 'w') as f:
        json.dump(settings, f)
    res = subprocess.run("cd '{}' && guniflask init".format(proj_dir),
                         shell=True)
    assert res.returncode == 0


def run_tests_of_project(proj_dir):
    res = subprocess.run("cd '{}' && pytest tests".format(proj_dir, proj_dir),
                         shell=True)
    assert res.returncode == 0


def run_debug_of_project(proj_dir):
    res = subprocess.run("cd '{}' && . bin/app-config.sh && echo $GUNIFLASK_ID_STRING".format(proj_dir),
                         shell=True, stdout=subprocess.PIPE)
    assert res.returncode == 0
    id_string = res.stdout.decode().strip()
    pid_file = join(proj_dir, '.pid', 'foo-' + id_string + '.pid')
    t = ExceptionThread(target=stop_project, args=(pid_file, proj_dir))
    t.start()
    res = subprocess.run("cd '{}' && bash bin/manage debug".format(proj_dir), shell=True)
    assert res.returncode == 0
    t.join()
    assert len(t.bucket) == 0, 'Exception in thread'


def run_table2model_without_db(proj_dir):
    res = subprocess.run("cd '{}' && bash bin/manage table2model".format(proj_dir), shell=True)
    assert res.returncode == 2


def run_initdb_without_db(proj_dir):
    res = subprocess.run("cd '{}' && bash bin/manage initdb -f".format(proj_dir), shell=True)
    assert res.returncode == 2


class ExceptionThread(Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bucket = []

    def run(self):
        try:
            super().run()
        except Exception as e:
            self.bucket.append(e)
            raise


def stop_project(pid_file, proj_dir):
    t = 10
    while t > 0 and not exists(pid_file):
        t -= 1
        time.sleep(1)
    assert t > 0, 'no project to stop'
    time.sleep(3)
    res = subprocess.run("cd '{}' && bash bin/manage stop".format(proj_dir), shell=True)
    assert res.returncode == 0
    t = 10
    while t > 0 and exists(pid_file):
        t -= 1
        time.sleep(1)
    assert t > 0, 'did not stop gracefully'


def test_init_project(tmpdir, monkeypatch):
    proj_dir = join(str(tmpdir), 'foo')
    os.mkdir(proj_dir)
    show_version()
    init_project(proj_dir)
    # run_tests_of_project(proj_dir)
    run_debug_of_project(proj_dir)
    run_initdb_without_db(proj_dir)
    run_table2model_without_db(proj_dir)



# coding=utf-8

import logging

from flask_sqlalchemy import SQLAlchemy

from guniflask.config import Config
from guniflask.security import JwtAuthManager

log = logging.getLogger(__name__)

config = Config()
db = SQLAlchemy()
jwt_manager = JwtAuthManager()


def make_settings(app, settings):
    """
    This function is invoked before initializing app.
    """


def init_app(app, settings):
    """
    This function is invoked before running app.
    """



# coding=utf-8

import os
from multiprocessing import Process
import signal
from threading import Thread
import time
from os.path import join
from logging.config import dictConfig
import logging

from gunicorn.glogging import Logger

from guniflask.app import create_bg_process_app


class BgProcess:
    def __init__(self, app):
        self.app = app
        self.settings = app.extensions['settings']
        self.configure_logging()

    def configure_logging(self):
        if 'bg_log_config' in self.settings:
            dictConfig(self.settings['bg_log_config'])
        else:
            logger = logging.getLogger(self.app.name)
            debug = os.environ.get('GUNIFLASK_DEBUG')
            if debug:
                logger.setLevel(logging.DEBUG)
            else:
                logger.setLevel(logging.INFO)
            log_dir = os.environ['GUNIFLASK_LOG_DIR']
            id_string = os.environ['GUNIFLASK_ID_STRING']
            project_name = os.environ.get('GUNIFLASK_PROJECT_NAME')
            log_file = join(log_dir, 'bg-{}-{}.log'.format(project_name, id_string))
            handler = logging.FileHandler(log_file)
            formatter = logging.Formatter(Logger.error_fmt, Logger.datefmt)
            handler.setFormatter(formatter)
            logger.addHandler(handler)

    def run(self):
        raise NotImplemented

    def start(self):
        with self.app.app_context():
            self.run()


def start_bg_process(server, name=None, bg_cls=None, on_starting=None):
    def start(pid):
        daemonize()
        init_signals()
        t = Thread(target=supervisor, args=(pid,))
        t.start()
        app = create_bg_process_app(name)
        bg = bg_cls(app)
        bg.start()

    def init_signals():
        def _exit(signum, frame):
            os._exit(0)

        signal.signal(signal.SIGINT, _exit)
        signal.signal(signal.SIGTERM, _exit)

    def supervisor(pid):
        while True:
            if not existsp(pid):
                os._exit(0)
            time.sleep(3)

    if on_starting is not None:
        on_starting(server)
    assert issubclass(bg_cls, BgProcess)
    pid = os.getpid()
    p = Process(target=start, args=(pid,))
    p.start()
    p.join()


def daemonize():
    if os.fork():
        os._exit(0)
    os.setsid()
    if os.fork():
        os._exit(0)
    os.umask(0o22)
    os.closerange(0, 3)
    fd_null = os.open(os.devnull, os.O_RDWR)
    if fd_null != 0:
        os.dup2(fd_null, 0)
    os.dup2(fd_null, 1)
    os.dup2(fd_null, 2)


def existsp(pid):
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True



# coding=utf-8

import datetime as dt
from functools import wraps

from flask import current_app, _request_ctx_stack, request, abort
from jwt import InvalidTokenError
from werkzeug.local import LocalProxy

from guniflask.utils.security import encode_jwt, decode_jwt

auth_manager = LocalProxy(lambda: current_app.extensions['auth_manager'])


def _load_user():
    ctx = _request_ctx_stack.top
    if ctx is not None:
        if not hasattr(ctx, 'user'):
            user = auth_manager.load_user()
            ctx.user = user
        return ctx.user


current_user = LocalProxy(_load_user)


class AuthManager:
    def __init__(self, app=None):
        self._user_loader = None
        self._unauthorized_handler = None
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.extensions['auth_manager'] = self

    def user_loader(self, func):
        self._user_loader = func
        return func

    def load_user(self):
        if self._user_loader is None:
            raise RuntimeError('Missing user loader')
        return self._user_loader()

    @property
    def unauthorized_view(self):
        if self._unauthorized_handler:
            return self._unauthorized_handler()
        return abort(401)

    def unauthorized(self, func):
        self._unauthorized_handler = func
        return func


class User:
    ROLE_PREFIX = 'role_'

    def __init__(self, authorities=None, **kwargs):
        self.authorities = []
        if authorities is not None:
            for a in authorities:
                self.authorities.append(a.lower())
        self._authorities_set = set(self.authorities)
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def is_authenticated(self):
        return True

    def has_authority(self, authority):
        authority = authority.lower()
        return authority in self._authorities_set

    def has_any_authority(self, *authorities):
        for a in authorities:
            if self.has_authority(a):
                return True
        return False

    def has_role(self, role):
        return self.has_authority(self._get_role_with_prefix(self.ROLE_PREFIX, role))

    def has_any_role(self, *roles):
        for r in roles:
            if self.has_role(r):
                return True
        return False

    @staticmethod
    def _get_role_with_prefix(prefix, role):
        if role is None or prefix is None:
            return role
        if role.startswith(prefix):
            return role
        return prefix + role


class AnonymousUser(User):
    def __init__(self, authorities=None, **kwargs):
        super().__init__(**kwargs)

    @property
    def is_authenticated(self):
        return False


def login_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not current_user.is_authenticated:
            return auth_manager.unauthorized_view
        return func(*args, **kwargs)

    return wrapper


def roles_required(*roles):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if current_user is None or not current_user.has_any_role(*roles):
                return auth_manager.unauthorized_view
            return func(*args, **kwargs)

        return wrapper

    return decorator


def authorities_required(*authorities):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if current_user is None or not current_user.has_any_authority(*authorities):
                return auth_manager.unauthorized_view
            return func(*args, **kwargs)

        return wrapper

    return decorator


class JwtAuthManager(AuthManager):
    def __init__(self, app=None):
        super().__init__(app=app)
        self._user_loader = self._load_user_from_header

    def init_app(self, app):
        self._set_default_app_config(app)
        super().init_app(app)

    def _set_default_app_config(self, app):
        app.config.setdefault('JWT_ALGORITHM', 'HS256')
        app.config.setdefault('JWT_SECRET', None)
        app.config.setdefault('JWT_PRIVATE_KEY', None)
        app.config.setdefault('JWT_PUBLIC_KEY', None)
        app.config.setdefault('JWT_ACCESS_TOKEN_EXPIRES_IN', dt.timedelta(days=1))

    def create_access_token(self, authorities=None, expires_in=None, **kwargs):
        config = current_app.config
        user_claims = dict(kwargs)
        user_claims['authorities'] = authorities
        payload = {'user_claims': user_claims}
        if expires_in is None:
            expires_in = config['JWT_ACCESS_TOKEN_EXPIRES_IN']
        return encode_jwt(payload, config['JWT_SECRET'] or config['JWT_PRIVATE_KEY'],
                          config['JWT_ALGORITHM'], expires_in=expires_in)

    def _load_user_from_header(self):
        config = current_app.config
        user = AnonymousUser()
        token = self.current_token
        if token:
            try:
                payload = decode_jwt(token, config['JWT_SECRET'] or config['JWT_PUBLIC_KEY'],
                                     config['JWT_ALGORITHM'])
            except InvalidTokenError:
                pass
            else:
                user_claims = payload['user_claims']
                user = User(**user_claims)
        return user

    @property
    def current_token(self):
        token = self._get_token_from_header()
        if token is None:
            token = self._get_token_from_query()
        return token

    @staticmethod
    def _get_token_from_header():
        auth = request.headers.get('Authorization')
        if auth is not None and auth.startswith('Bearer'):
            return auth.split(' ', 1)[1]

    @staticmethod
    def _get_token_from_query():
        return request.args.get('access_token')



# coding=utf-8

import os
from os.path import isfile, join, isdir
from importlib import import_module
from pkgutil import iter_modules


def load_config(fname, **kwargs):
    if fname is None or not isfile(fname):
        raise FileNotFoundError("Cannot find configuration file '{}'".format(fname))
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    cfg.update(kwargs)
    exec(code, cfg, cfg)
    return cfg


def load_profile_config(conf_dir, name, profiles=None, **kwargs):
    pc = load_config(join(conf_dir, name + '.py'), **kwargs)
    if profiles:
        profiles = profiles.split(',')
        profiles.reverse()
        for profile in profiles:
            if profile:
                pc_file = join(conf_dir, name + '_' + profile + '.py')
                if isfile(pc_file):
                    c = load_config(pc_file, **kwargs)
                    pc.update(c)
    return pc


def walk_modules(path):
    mods = []
    mod = import_module(path)
    mods.append(mod)
    if hasattr(mod, '__path__'):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + '.' + subpath
            if ispkg:
                mods += walk_modules(fullpath)
            else:
                submod = import_module(fullpath)
                mods.append(submod)
    return mods


def walk_files(path):
    files = []
    if isdir(path):
        names = os.listdir(path)
        for name in names:
            files += walk_files(join(path, name))
    elif isfile(path):
        files.append(path)
    return files




# coding=utf-8

import logging
from importlib import import_module

from flask import Flask, Blueprint
from flask_cors import CORS
from flask_sqlalchemy import SQLAlchemy

from guniflask.config import Config
from guniflask.utils.config import walk_modules
from guniflask.utils.logging import redirect_app_logger, redirect_logger
from guniflask.model import wrap_model
from guniflask.security import JwtAuthManager
from guniflask.apidoc import ApiDoc

log = logging.getLogger(__name__)

app_default_settings = {
    'debug': False,
    'cors': True,
    # Flask-SQLAlchemy
    'SQLALCHEMY_TRACK_MODIFICATIONS': False
}


def _get_app_module(app):
    return import_module(app.name + '.app')


def _get_instance_from_app(app, name):
    app_module = _get_app_module(app)
    obj = getattr(app_module, name, None)
    return obj


def set_app_config(app):
    app_module = _get_app_module(app)
    config = _get_instance_from_app(app, 'config')
    assert isinstance(config, Config)

    config.init_app(app)
    s = config.app_settings(app)
    for k, v in app_default_settings.items():
        s.setdefault(k, v)

    _make_settings = getattr(app_module, 'make_settings', None)
    if _make_settings:
        _make_settings(app, s)

    for k, v in s.items():
        if k.isupper():
            app.config[k] = v

    log.info('%s active profiles: %s', app.name, s['active_profiles'])


def init_app(app):
    app_module = _get_app_module(app)
    config = _get_instance_from_app(app, 'config')
    assert isinstance(config, Config)

    s = config.app_settings(app)

    # CORS
    if s['cors']:
        if isinstance(s['cors'], dict):
            CORS(app, **s['cors'])
        else:
            CORS(app)

    # database configuration
    if 'SQLALCHEMY_DATABASE_URI' in app.config:
        db = _get_instance_from_app(app, 'db')
        assert isinstance(db, SQLAlchemy)
        wrap_model(db.Model)
        db.init_app(app)

    # authentication
    jwt_manager = _get_instance_from_app(app, 'jwt_manager')
    if isinstance(jwt_manager, JwtAuthManager):
        jwt_manager.init_app(app)

    # API doc
    if s['apidoc'] or (s['apidoc'] is None and s['debug']):
        apidoc = ApiDoc()
        apidoc.init_app(app)

    _init_app = getattr(app_module, 'init_app', None)
    if _init_app:
        _init_app(app, s)


def register_blueprints(app):
    def iter_blueprints():
        for module in walk_modules(app.name):
            for obj in vars(module).values():
                if isinstance(obj, Blueprint):
                    yield obj

    for b in iter_blueprints():
        app.register_blueprint(b)


def create_app(name):
    app = Flask(name)
    gunicorn_logger = logging.getLogger('gunicorn.error')
    redirect_app_logger(app, gunicorn_logger)
    redirect_logger(name, gunicorn_logger)
    register_blueprints(app)
    set_app_config(app)
    init_app(app)
    return app


def create_bg_process_app(name):
    app = Flask(name)
    set_app_config(app)
    init_app(app)
    return app



# coding=utf-8

import re
from os.path import dirname, join

from jinja2 import Environment

from guniflask.errors import TemplateError

template_folder = join(dirname(dirname(__file__)), 'templates')

_camelcase_invalid_chars = re.compile(r'[^a-zA-Z\d]')

_name_invalid_chars = re.compile(r'[^a-zA-Z\d\-_]')


def string_camelcase(s):
    return _camelcase_invalid_chars.sub('', s.title())


def string_lowercase_hyphen(s):
    return _name_invalid_chars.sub('', s.strip().replace(' ', '-').lower()).replace('_', '-')


def string_lowercase_underscore(s):
    return _name_invalid_chars.sub('', s.strip().replace(' ', '_').lower()).replace('-', '_')


def string_uppercase_underscore(s):
    return _name_invalid_chars.sub('', s.strip().replace(' ', '_').upper()).replace('-', '_')


def _raise_helper(message):
    if message:
        raise TemplateError(message)
    raise TemplateError


def _assert_helper(logical, message=None):
    if not logical:
        _raise_helper(message)
    return ''


def jinja2_env():
    env = Environment(keep_trailing_newline=True)
    env.globals['raise'] = _raise_helper
    env.globals['assert'] = _assert_helper
    return env



# coding=utf-8

import os
import datetime as dt
import base64
import uuid

import jwt


def generate_jwt_secret(n=32):
    return base64.b64encode(os.urandom(n), altchars=b'-_').decode().replace('=', '')


def decode_jwt(token, key, algorithm, **kwargs):
    return jwt.decode(token, key=key, algorithms=[algorithm], **kwargs)


def encode_jwt(payload, key, algorithm, expires_in=None, **kwargs):
    token_data = {
        'jti': str(uuid.uuid4())
    }
    if expires_in is not None:
        if isinstance(expires_in, (int, float)):
            expires_in = dt.timedelta(seconds=expires_in)
        exp = dt.datetime.utcnow() + expires_in
        token_data['exp'] = exp
    token_data.update(payload)
    token = jwt.encode(token_data, key, algorithm=algorithm, **kwargs)
    if isinstance(token, bytes):
        token = token.decode('utf-8')
    return token



# coding=utf-8

import os
import copy
from collections import MutableMapping

from flask import current_app
from werkzeug.local import LocalProxy

from guniflask.utils.config import load_profile_config

settings = LocalProxy(lambda: current_app.extensions['settings'])


class Config:
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        s = Settings(self._load_app_settings(app))
        app.extensions['settings'] = s

    def _load_app_settings(self, app):
        c = {}
        conf_dir = os.environ.get('GUNIFLASK_CONF_DIR')
        active_profiles = os.environ.get('GUNIFLASK_ACTIVE_PROFILES')
        kwargs = get_default_args_from_env()
        if conf_dir:
            c = load_profile_config(conf_dir, app.name, profiles=active_profiles, **kwargs)
            c['active_profiles'] = active_profiles
        c.update(kwargs)
        s = {}
        for name in c:
            if not name.startswith('_'):
                s[name] = c[name]
        return s

    @property
    def settings(self):
        return self.app_settings()

    def app_settings(self, app=None):
        if app is None:
            return current_app.extensions['settings']
        return app.extensions['settings']


def get_default_args_from_env():
    kwargs = {'home': os.environ.get('GUNIFLASK_HOME', os.curdir),
              'project_name': os.environ.get('GUNIFLASK_PROJECT_NAME')}
    if os.environ.get('GUNIFLASK_DEBUG'):
        kwargs['debug'] = True
    else:
        kwargs['debug'] = False
    return kwargs


class Settings(MutableMapping):
    def __init__(self, __values=None, **kwargs):
        self.attributes = {}
        self.update(__values, **kwargs)

    def __getitem__(self, name):
        if name not in self:
            return None
        return self.attributes[name]

    def __contains__(self, name):
        return name in self.attributes

    def get(self, name, default=None):
        return self[name] if self[name] is not None else default

    def getbool(self, name, default=None):
        v = self.get(name, default)
        return getbool(v)

    def getint(self, name, default=None):
        v = self.get(name, default)
        return getint(v)

    def getfloat(self, name, default=None):
        v = self.get(name, default)
        return getfloat(v)

    def getlist(self, name, default=None):
        v = self.get(name, default)
        return getlist(v)

    def __setitem__(self, name, value):
        self.attributes[name] = value

    def set(self, name, value):
        self[name] = value

    def setdefault(self, k, default=None):
        if k not in self:
            self[k] = default

    def update(self, __values=None, **kwargs):
        if __values is not None:
            if isinstance(__values, Settings):
                for name in __values:
                    self[name] = __values[name]
            else:
                for name, value in __values.items():
                    self[name] = value
        for k, v in kwargs.items():
            self[k] = v

    def delete(self, name):
        del self.attributes[name]

    def __delitem__(self, name):
        del self.attributes[name]

    def copy(self):
        return copy.deepcopy(self)

    def __iter__(self):
        return iter(self.attributes)

    def __len__(self):
        return len(self.attributes)


def getbool(v):
    try:
        return bool(int(v))
    except (ValueError, TypeError):
        if v in ("True", "true"):
            return True
        if v in ("False", "false"):
            return False
    return None


def getint(v):
    try:
        return int(v)
    except (ValueError, TypeError):
        pass
    return None


def getfloat(v):
    try:
        return float(v)
    except (ValueError, TypeError):
        pass
    return None


def getlist(v):
    if v is None:
        return None
    if isinstance(v, str):
        v = v.split(",")
    elif not hasattr(v, "__iter__"):
        v = [v]
    return list(v)



# coding=utf-8

import sys
import os
from os.path import isfile, join, dirname, exists, isdir
import argparse
import signal
import time
import multiprocessing
from importlib import import_module
import json
from functools import partial

from sqlalchemy.engine import create_engine
from sqlalchemy.schema import MetaData

from gunicorn.config import KNOWN_SETTINGS
from gunicorn.app.base import Application

from guniflask.utils.config import walk_modules, load_profile_config, walk_files
from guniflask.modelgen import SqlToModelGenerator
from guniflask.errors import UsageError
from guniflask.commands import Command
from guniflask.app import create_app
from guniflask.bg_process import start_bg_process


def set_environ():
    home_dir = os.environ.get('GUNIFLASK_HOME')
    if home_dir is None:
        home_dir = os.getcwd()
        os.environ['GUNIFLASK_HOME'] = home_dir
    if home_dir not in sys.path:
        sys.path.append(home_dir)
    if 'GUNIFLASK_PROJECT_NAME' not in os.environ:
        project_name = _infer_project_name(home_dir)
        if project_name:
            os.environ['GUNIFLASK_PROJECT_NAME'] = project_name
    if 'GUNIFLASK_CONF_DIR' not in os.environ:
        os.environ['GUNIFLASK_CONF_DIR'] = join(home_dir, 'conf')
    if 'GUNIFLASK_LOG_DIR' not in os.environ:
        os.environ['GUNIFLASK_LOG_DIR'] = join(home_dir, '.log')
    if 'GUNIFLASK_PID_DIR' not in os.environ:
        os.environ['GUNIFLASK_PID_DIR'] = join(home_dir, '.pid')
    if 'GUNIFLASK_ID_STRING' not in os.environ:
        os.environ['GUNIFLASK_ID_STRING'] = os.getlogin()


def _get_project_name():
    return os.environ.get('GUNIFLASK_PROJECT_NAME')


def _infer_project_name(home_dir):
    init_file = join(home_dir, '.guniflask-init.json')
    if isfile(init_file):
        try:
            with open(init_file, 'r') as f:
                data = json.load(f)
            project_name = data.get('project_name')
            if project_name and isdir(join(home_dir, project_name)):
                return project_name
        except Exception:
            pass
    for d in os.listdir(home_dir):
        if not d.startswith('.') and isfile(join(home_dir, d, 'app.py')):
            return d


class InitDb(Command):
    @property
    def name(self):
        return 'initdb'

    @property
    def syntax(self):
        return '[options]'

    @property
    def short_desc(self):
        return 'Initialize database from definition of models'

    def add_arguments(self, parser):
        parser.add_argument('-f', '--force', dest='force', action='store_true', default=False,
                            help='force creating all tables')

    def run(self, args):
        project_name = _get_project_name()
        project_module = import_module(project_name)
        db = getattr(project_module, 'db')
        settings = getattr(project_module, 'settings')

        walk_modules(project_name)
        app = create_app(project_name)
        with app.app_context():
            if settings.get('SQLALCHEMY_DATABASE_URI') is None:
                raise UsageError("Please set 'SQLALCHEMY_DATABASE_URI' in configuration.")
            if args.force:
                db.drop_all()
            else:
                print("\033[33mThe tables already exist will be skipped.\033[0m")
                print("\033[33mYou can try '-f' option to force creating all tables.\033[0m")
            db.create_all()


class TableToModel(Command):
    @property
    def name(self):
        return 'table2model'

    @property
    def syntax(self):
        return '[options]'

    @property
    def short_desc(self):
        return 'Convert database tables to definition of models'

    def add_arguments(self, parser):
        parser.add_argument('--tables', dest='tables',
                            help='tables to process (comma-separated, default: all)')

    def process_arguments(self, args):
        if args.tables is not None:
            args.tables = args.tables.split(',')

    def run(self, args):
        project_name = _get_project_name()
        project_module = import_module(project_name)
        settings = getattr(project_module, 'settings')

        app = create_app(project_name)
        with app.app_context():
            database_uri = settings.get('SQLALCHEMY_DATABASE_URI')
            if database_uri is None:
                raise UsageError("Please set 'SQLALCHEMY_DATABASE_URI' in configuration.")
            engine = create_engine(database_uri)
            metadata = MetaData(engine)
            metadata.reflect(only=args.tables)
            gen = SqlToModelGenerator(project_name, metadata)
            dest = settings.get('table2model_dest', join(project_name, 'models'))
            gen.render(join(settings['home'], dest))


class Debug(Command):
    @property
    def name(self):
        return 'debug'

    @property
    def short_desc(self):
        return 'Debug application'

    def add_arguments(self, parser):
        parser.add_argument('-d', '--daemon', dest='daemon', action='store_true', help='run in daemon mode')
        parser.add_argument('-p', '--active-profiles', dest='active_profiles', metavar='PROFILES',
                            help='active profiles (comma-separated)')

    def process_arguments(self, args):
        if args.active_profiles:
            os.environ['GUNIFLASK_ACTIVE_PROFILES'] = args.active_profiles

    def run(self, args):
        os.environ['GUNIFLASK_DEBUG'] = '1'
        os.environ.setdefault('GUNIFLASK_ACTIVE_PROFILES', 'dev')
        app = GunicornApplication()
        if args.daemon:
            app.set_option('daemon', True)
        pid = get_pid(app.options)
        if pid is not None and is_started(pid):
            print('Application is already started')
            self.exitcode = 1
        else:
            app.run()


class Start(Command):
    @property
    def name(self):
        return 'start'

    @property
    def short_desc(self):
        return 'Start application'

    def add_arguments(self, parser):
        parser.add_argument('--daemon-off', dest='daemon_off', action='store_true', help='turn off daemon mode')
        parser.add_argument('-p', '--active-profiles', dest='active_profiles', metavar='PROFILES',
                            help='active profiles (comma-separated)')

    def process_arguments(self, args):
        if args.active_profiles:
            os.environ['GUNIFLASK_ACTIVE_PROFILES'] = args.active_profiles

    def run(self, args):
        os.environ.setdefault('GUNIFLASK_ACTIVE_PROFILES', 'prod')
        app = GunicornApplication()
        if args.daemon_off:
            app.set_option('daemon', False)
        pid = get_pid(app.options)
        if pid is not None and is_started(pid):
            print('Application is already started')
            self.exitcode = 1
        else:
            app.run()


class Stop(Command):
    @property
    def name(self):
        return 'stop'

    @property
    def short_desc(self):
        return 'Stop application'

    def run(self, args):
        app = GunicornApplication()
        pid = get_pid(app.options)
        if pid is None or not is_started(pid):
            print('No application to stop')
            self.exitcode = 1
        else:
            print('kill {}'.format(pid))
            os.kill(pid, signal.SIGTERM)
            time.sleep(3)
            try:
                os.kill(pid, 0)
            except OSError:
                pass
            else:
                print('Application did not stop gracefully after 3 seconds')
                print('kill -9 {}'.format(pid))
                os.kill(pid, signal.SIGKILL)


class GunicornApplication(Application):
    known_settings = {'bg_process'}

    def __init__(self):
        self.options = self._make_options()
        super().__init__()

    def set_option(self, key, value):
        if key in self.cfg.settings:
            self.cfg.set(key, value)

    def load_config(self):
        for key, value in self.options.items():
            if key in self.cfg.settings and value is not None:
                self.cfg.set(key.lower(), value)

    def load(self):
        return create_app(_get_project_name())

    def _make_options(self):
        pid_dir = os.environ['GUNIFLASK_PID_DIR']
        log_dir = os.environ['GUNIFLASK_LOG_DIR']
        id_string = os.environ['GUNIFLASK_ID_STRING']
        project_name = _get_project_name()
        options = {
            'daemon': True,
            'preload_app': True,
            'workers': multiprocessing.cpu_count(),
            'worker_class': 'gevent',
            'pidfile': join(pid_dir, '{}-{}.pid'.format(project_name, id_string)),
            'accesslog': join(log_dir, '{}-{}.access.log'.format(project_name, id_string)),
            'errorlog': join(log_dir, '{}-{}.error.log'.format(project_name, id_string))
        }
        options.update(self._make_profile_options(os.environ.get('GUNIFLASK_ACTIVE_PROFILES')))
        # if debug
        if os.environ.get('GUNIFLASK_DEBUG'):
            options.update(self._make_debug_options())
        self._makedirs(options)
        # bg process
        self._set_bg_process(options)
        return options

    def _set_bg_process(self, options):
        if 'bg_process' in options:
            bg_cls = options['bg_process']
            kwargs = dict(name=_get_project_name(), bg_cls=bg_cls, on_starting=options.get('on_starting'))
            options['on_starting'] = partial(start_bg_process, **kwargs)
            options.pop('bg_process')

    def _make_profile_options(self, active_profiles):
        conf_dir = os.environ['GUNIFLASK_CONF_DIR']
        gc = load_profile_config(conf_dir, 'gunicorn', profiles=active_profiles)
        settings = {}
        snames = set([i.name for i in KNOWN_SETTINGS])
        snames.update(self.known_settings)
        for name in gc:
            if name in snames:
                settings[name] = gc[name]
        return settings

    @staticmethod
    def _make_debug_options():
        conf_dir = os.environ['GUNIFLASK_CONF_DIR']
        return {
            'accesslog': '-',
            'errorlog': '-',
            'loglevel': 'debug',
            'disable_redirect_access_to_syslog': True,
            'preload_app': False,
            'reload': True,
            'reload_extra_files': walk_files(conf_dir),
            'workers': 1,
            'daemon': False
        }

    @staticmethod
    def _makedirs(opts):
        for c in ['pidfile', 'accesslog', 'errorlog']:
            p = opts.get(c)
            if p:
                d = dirname(p)
                if d and not exists(d):
                    os.makedirs(d)


def get_pid(options):
    pidfile = options.get('pidfile')
    if isfile(pidfile):
        with open(pidfile, 'r') as f:
            line = f.readline()
            if line:
                pid = line.strip()
                if pid:
                    return int(pid)


def is_started(pid):
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True


def _get_commands_from_module():
    d = {}
    for cmd in globals().values():
        if isinstance(cmd, type) and issubclass(cmd, Command):
            o = cmd()
            if o.name:
                d[o.name] = o
    return d


def _print_commands():
    print("Usage: guniflask-manage <command> [options] [args]\n")
    print("Available commands:")
    cmds = _get_commands_from_module()
    for cmdname, cmdclass in sorted(cmds.items()):
        print("   {:<13} {}".format(cmdname, cmdclass.short_desc))
    print()
    print('Use "guniflask-manage <command> -h" to see more info about a command')


def _print_unknown_command(cmdname):
    print("Unknown command: %s\n" % cmdname, file=sys.stderr)
    print('Use "guniflask-manage" to see available commands', file=sys.stderr)


def main(argv=None):
    set_environ()
    if argv is None:
        argv = sys.argv
    cmds = _get_commands_from_module()
    cmdname = argv[1] if len(argv) > 1 else None
    if not cmdname or cmdname in ('-h', '--help'):
        _print_commands()
        sys.exit(0)
    elif cmdname not in cmds:
        _print_unknown_command(cmdname)
        sys.exit(2)
    del argv[1]
    cmd = cmds[cmdname]
    parser = argparse.ArgumentParser()
    parser.usage = "manage {} {}".format(cmdname, cmd.syntax)
    parser.description = cmd.long_desc
    cmd.add_arguments(parser)
    try:
        args = parser.parse_args(args=argv[1:])
        cmd.process_arguments(args)
        cmd.run(args)
    except UsageError as e:
        print('Error: {}'.format(e), file=sys.stderr)
        sys.exit(2)
    else:
        if cmd.exitcode:
            sys.exit(cmd.exitcode)


if __name__ == '__main__':
    main()



# coding=utf-8

import datetime as dt

import sqlalchemy


def model_to_dict(model, ignore=None):
    col_attrs = sqlalchemy.inspect(model).mapper.column_attrs
    d = {}
    tz_info = dt.datetime.now(tz=dt.timezone.utc).astimezone().tzinfo
    ignore_set = _get_ignore_set(ignore)
    for c in col_attrs:
        if c.key not in ignore_set:
            v = getattr(model, c.key)
            if isinstance(v, dt.datetime) and v.tzinfo is None:
                v = v.replace(tzinfo=tz_info)
            d[c.key] = v
    return d


def dict_to_model(dict_obj, model_cls=None, ignore=None):
    if model_cls is None:
        return dict(dict_obj)
    mapper = sqlalchemy.inspect(model_cls).mapper
    columns = mapper.columns
    col_attrs = mapper.column_attrs
    ignore_set = _get_ignore_set(ignore)
    kwargs = {}
    for c in col_attrs:
        if c.key in dict_obj and c.key not in ignore_set:
            v = dict_obj[c.key]
            if isinstance(columns[c.key].type, sqlalchemy.DateTime):
                v = string_to_datetime(v)
            kwargs[c.key] = v
    model = model_cls(**kwargs)
    return model


def update_model_by_dict(model, dict_obj, ignore=None):
    mapper = sqlalchemy.inspect(model).mapper
    columns = mapper.columns
    col_attrs = mapper.column_attrs
    ignore_set = _get_ignore_set(ignore)
    for c in col_attrs:
        if c.key in dict_obj and c.key not in ignore_set:
            v = dict_obj[c.key]
            if isinstance(columns[c.key].type, sqlalchemy.DateTime):
                v = string_to_datetime(v)
            setattr(model, c.key, v)


def string_to_datetime(s):
    if isinstance(s, int):
        return dt.datetime.fromtimestamp(s, tz=dt.timezone.utc).astimezone()
    elif isinstance(s, str):
        if 'GMT' in s:
            return dt.datetime.strptime(s, '%a, %d %b %Y %H:%M:%S GMT').replace(tzinfo=dt.timezone.utc).astimezone()
    return s


def _get_ignore_set(ignore):
    if isinstance(ignore, str):
        ignore = [i.strip() for i in ignore.split(',')]
    return set(ignore if ignore else [])


def wrap_model(model_cls):
    if not hasattr(model_cls, 'to_dict'):
        model_cls.to_dict = lambda self, **kwargs: model_to_dict(self, **kwargs)
    if not hasattr(model_cls, 'from_dict'):
        model_cls.from_dict = classmethod(lambda cls, dict_obj, **kwargs:
                                          dict_to_model(dict_obj, model_cls=cls, **kwargs))
    if not hasattr(model_cls, 'update_by_dict'):
        model_cls.update_by_dict = lambda self, dict_obj, **kwargs: update_model_by_dict(self, dict_obj, **kwargs)
    return model_cls
